package ru.dbotthepony.kstarbound.defs

import com.google.common.collect.ImmutableSet
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.builder.JsonImplementation
import ru.dbotthepony.kstarbound.util.KOptional

val BaseMovementParameters.ignorePlatformCollision get() = if (this is MovementParameters) this.ignorePlatformCollision ?: false else false
val BaseMovementParameters.restDuration get() = if (this is MovementParameters) this.restDuration ?: 0 else 0

@JsonImplementation(BaseMovementParameters.Impl::class)
sealed interface BaseMovementParameters {
	val mass: Double?
	val gravityMultiplier: Double?
	val liquidBuoyancy: Double?
	val airBuoyancy: Double?
	val bounceFactor: Double?

	// If set to true, during an update that has more than one internal movement
	// step, the movement will stop on the first bounce.
	val stopOnFirstBounce: Boolean?

	// Cheat when sliding on the ground, by trying to correct upwards before
	// other directions (within a set limit).  Allows smooth sliding along
	// horizontal ground without losing horizontal speed.
	val enableSurfaceSlopeCorrection: Boolean?
	val slopeSlidingFactor: Double?

	// ignored
	val maxMovementPerStep: Double?
	val maximumCorrection: Double?
	val speedLimit: Double?

	val stickyCollision: Boolean?
	val stickyForce: Double?

	val airFriction: Double?
	val liquidFriction: Double?
	val groundFriction: Double?

	val collisionEnabled: Boolean?
	val frictionEnabled: Boolean?
	val gravityEnabled: Boolean?

	val maximumPlatformCorrection: Double?
	val maximumPlatformCorrectionVelocityFactor: Double?

	val physicsEffectCategories: ImmutableSet<String>?

	@JsonFactory
	data class Impl(
		override val mass: Double? = null,
		override val gravityMultiplier: Double? = null,
		override val liquidBuoyancy: Double? = null,
		override val airBuoyancy: Double? = null,
		override val bounceFactor: Double? = null,
		override val stopOnFirstBounce: Boolean? = null,
		override val enableSurfaceSlopeCorrection: Boolean? = null,
		override val slopeSlidingFactor: Double? = null,
		override val maxMovementPerStep: Double? = null,
		override val maximumCorrection: Double? = null,
		override val speedLimit: Double? = null,
		override val stickyCollision: Boolean? = null,
		override val stickyForce: Double? = null,
		override val airFriction: Double? = null,
		override val liquidFriction: Double? = null,
		override val groundFriction: Double? = null,
		override val collisionEnabled: Boolean? = null,
		override val frictionEnabled: Boolean? = null,
		override val gravityEnabled: Boolean? = null,
		override val maximumPlatformCorrection: Double? = null,
		override val maximumPlatformCorrectionVelocityFactor: Double? = null,
		override val physicsEffectCategories: ImmutableSet<String>? = null,
	) : BaseMovementParameters {
		fun merge(other: Impl): Impl {
			return Impl(
				mass = mass ?: other.mass,
				gravityMultiplier = gravityMultiplier ?: other.gravityMultiplier,
				liquidBuoyancy = liquidBuoyancy ?: other.liquidBuoyancy,
				airBuoyancy = airBuoyancy ?: other.airBuoyancy,
				bounceFactor = bounceFactor ?: other.bounceFactor,
				stopOnFirstBounce = stopOnFirstBounce ?: other.stopOnFirstBounce,
				enableSurfaceSlopeCorrection = enableSurfaceSlopeCorrection ?: other.enableSurfaceSlopeCorrection,
				slopeSlidingFactor = slopeSlidingFactor ?: other.slopeSlidingFactor,
				maxMovementPerStep = maxMovementPerStep ?: other.maxMovementPerStep,
				maximumCorrection = maximumCorrection ?: other.maximumCorrection,
				speedLimit = speedLimit ?: other.speedLimit,
				stickyCollision = stickyCollision ?: other.stickyCollision,
				stickyForce = stickyForce ?: other.stickyForce,
				airFriction = airFriction ?: other.airFriction,
				liquidFriction = liquidFriction ?: other.liquidFriction,
				groundFriction = groundFriction ?: other.groundFriction,
				collisionEnabled = collisionEnabled ?: other.collisionEnabled,
				frictionEnabled = frictionEnabled ?: other.frictionEnabled,
				gravityEnabled = gravityEnabled ?: other.gravityEnabled,
				maximumPlatformCorrection = maximumPlatformCorrection ?: other.maximumPlatformCorrection,
				maximumPlatformCorrectionVelocityFactor = maximumPlatformCorrectionVelocityFactor ?: other.maximumPlatformCorrectionVelocityFactor,
				physicsEffectCategories = physicsEffectCategories ?: other.physicsEffectCategories,
			)
		}
	}
}