package ru.dbotthepony.kstarbound.defs

import ru.dbotthepony.kstarbound.json.builder.JsonAlias
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.builder.JsonFlat
import ru.dbotthepony.kstarbound.world.physics.Poly

@JsonFactory
data class PlayerMovementParameters(
	@JsonFlat
	val base: BaseMovementParameters.Impl = BaseMovementParameters.Impl(),

	@JsonAlias("collisionPoly")
	val standingPoly: Poly? = null,
	@JsonAlias("collisionPoly")
	val crouchingPoly: Poly? = null,

	val walkSpeed: Double? = null,
	val runSpeed: Double? = null,
	val flySpeed: Double? = null,

	val minimumLiquidPercentage: Double? = null,
	val liquidImpedance: Double? = null,
	val normalGroundFriction: Double? = null,
	val ambulatingGroundFriction: Double? = null,
	val groundForce: Double? = null,
	val airForce: Double? = null,
	val liquidForce: Double? = null,

	val airJumpProfile: JumpProfile = JumpProfile(),
	val liquidJumpProfile: JumpProfile = JumpProfile(),

	val fallStatusSpeedMin: Double? = null,
	val fallThroughSustainFrames: Int? = null,

	val groundMovementMinimumSustain: Double? = null,
	val groundMovementMaximumSustain: Double? = null,
	val groundMovementCheckDistance: Double? = null,

	val pathExploreRate: Double? = null,
) : BaseMovementParameters by base {
	fun merge(other: PlayerMovementParameters): PlayerMovementParameters {
		return PlayerMovementParameters(
			base = base.merge(other.base),
			standingPoly = standingPoly ?: other.standingPoly,
			crouchingPoly = crouchingPoly ?: other.crouchingPoly,
			walkSpeed = walkSpeed ?: other.walkSpeed,
			runSpeed = runSpeed ?: other.runSpeed,
			flySpeed = flySpeed ?: other.flySpeed,
			minimumLiquidPercentage = minimumLiquidPercentage ?: other.minimumLiquidPercentage,
			liquidImpedance = liquidImpedance ?: other.liquidImpedance,
			normalGroundFriction = normalGroundFriction ?: other.normalGroundFriction,
			ambulatingGroundFriction = ambulatingGroundFriction ?: other.ambulatingGroundFriction,
			groundForce = groundForce ?: other.groundForce,
			airForce = airForce ?: other.airForce,
			liquidForce = liquidForce ?: other.liquidForce,
			airJumpProfile = airJumpProfile.merge(other.airJumpProfile),
			liquidJumpProfile = liquidJumpProfile.merge(other.liquidJumpProfile),
			fallStatusSpeedMin = fallStatusSpeedMin ?: other.fallStatusSpeedMin,
			fallThroughSustainFrames = fallThroughSustainFrames ?: other.fallThroughSustainFrames,
			groundMovementMinimumSustain = groundMovementMinimumSustain ?: other.groundMovementMinimumSustain,
			groundMovementMaximumSustain = groundMovementMaximumSustain ?: other.groundMovementMaximumSustain,
			groundMovementCheckDistance = groundMovementCheckDistance ?: other.groundMovementCheckDistance,
		)
	}
}
