package ru.dbotthepony.kstarbound.world.api

import com.github.benmanes.caffeine.cache.Interner
import ru.dbotthepony.kstarbound.util.HashTableInterner
import java.io.DataInputStream

sealed class AbstractCell {
	abstract val foreground: AbstractTileState
	abstract val background: AbstractTileState
	abstract val liquid: AbstractLiquidState

	abstract val dungeonId: Int
	abstract val biome: Int
	abstract val envBiome: Int
	abstract val isIndestructible: Boolean

	abstract fun immutable(): ImmutableCell
	abstract fun mutable(): MutableCell

	companion object {
		fun skip(stream: DataInputStream) {
			AbstractTileState.skip(stream)
			AbstractTileState.skip(stream)
			AbstractLiquidState.skip(stream)
			stream.skipNBytes(1 + 2 + 1 + 1 + 1 + 1)
		}

		@JvmStatic
		protected val POOL: Interner<ImmutableCell> = HashTableInterner()

		val EMPTY: ImmutableCell = POOL.intern(ImmutableCell(AbstractTileState.EMPTY, AbstractTileState.EMPTY, AbstractLiquidState.EMPTY, 0, 0, 0, false))
		val NULL: ImmutableCell = POOL.intern(ImmutableCell(AbstractTileState.NULL, AbstractTileState.NULL, AbstractLiquidState.EMPTY, 0, 0, 0, false))
	}
}
