package ru.dbotthepony.kstarbound.world.api

import com.github.benmanes.caffeine.cache.Interner
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.defs.tile.BuiltinMetaMaterials
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.util.HashTableInterner
import java.io.DataInputStream

sealed class AbstractTileState {
	abstract val material: Registry.Entry<TileDefinition>
	abstract val modifier: Registry.Entry<MaterialModifier>?
	abstract val color: TileColor
	abstract val hueShift: Float
	abstract val modifierHueShift: Float

	abstract fun immutable(): ImmutableTileState
	abstract fun mutable(): MutableTileState

	companion object {
		fun skip(stream: DataInputStream) {
			stream.skipNBytes(2 + 1 + 1 + 2 + 1)
		}

		@JvmStatic
		protected val POOL: Interner<ImmutableTileState> = HashTableInterner()

		val EMPTY: ImmutableTileState = POOL.intern(ImmutableTileState(BuiltinMetaMaterials.EMPTY))
		val NULL: ImmutableTileState = POOL.intern(ImmutableTileState(BuiltinMetaMaterials.NULL))
	}
}
